from typing import Tuple

import tensorflow as tf

from ptcn.components import TemporalResidualBlock, PERMITTED_PADDING_TYPES
from ptcn.constants import DEFAULT_RANDOM_SEED


class GenericTemporalConvNet(tf.keras.Model):
    problem_type_activations = {
        "generic": "relu",
        "regression": "linear",
        "classification": "sigmoid"
    }

    def __init__(
            self,
            n_filters: Tuple[int, ...],
            kernel_size: int = 3,
            padding_type: str = 'causal',
            dropout_rate: float = 0.2,
            problem_type: str = 'generic'
    ):
        """A generic Temporal Convolutional Network

        Expects data to be provided in the following shape: (batch_size, n_timesteps, n_features)

        :param n_filters:
        :param kernel_size:
        :param padding_type:
        :param dropout_rate:
        """
        super(GenericTemporalConvNet, self).__init__()
        # validate args
        assert padding_type in PERMITTED_PADDING_TYPES, f"Value of 'padding_type' parameter must be one of {PERMITTED_PADDING_TYPES}"
        assert 0. < dropout_rate < 1., r"Value of 'dropout_rate' parameter must lie inside the range (0., 1.)"
        # TODO: validate n_filters
        assert problem_type in ['generic', 'regression', 'classification']

        # Create network
        self.network = tf.keras.Sequential(
            [
                self._initialise_res_block(
                    dropout_rate,
                    kernel_size,
                    layer_filters,
                    layer_idx,
                    n_filters,
                    padding_type,
                    problem_type,
                )
                for layer_idx, layer_filters in enumerate(n_filters)
            ]
        )

    def _initialise_res_block(self, dropout_rate, kernel_size, layer_filters, layer_idx, n_filters, padding_type,
                              problem_type):
        layer_activation = self._get_activation(layer_idx, n_filters, problem_type)
        return TemporalResidualBlock(
            dilation=2 ** layer_idx,
            n_filters=layer_filters,
            kernel_size=kernel_size,
            padding_type=padding_type,
            dropout_rate=dropout_rate,
            final_activation=layer_activation,
            random_seed=DEFAULT_RANDOM_SEED * (layer_idx + 1),
            name=f"resblock_{layer_idx}"
        )

    def _get_activation(self, layer_idx: int, n_filters: Tuple, problem_type: str):
        return self.problem_type_activations.get(problem_type) if layer_idx + 1 == len(n_filters) else 'relu'

    def call(self, inputs, training=None, mask=None):
        return self.network(inputs, training=training)



import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as layers
import tensorflow_addons as tfa

from ptcn.constants import DEFAULT_RANDOM_SEED

PERMITTED_PADDING_TYPES = ['valid', 'causal', 'same']


class TemporalResidualBlock(tf.keras.layers.Layer):

    def __init__(
            self,
            dilation: int,
            n_filters: int,
            kernel_size: int = 3,
            padding_type: str = 'causal',
            dropout_rate: float = 0.2,
            final_activation: str = 'relu',
            name: str = 'TemporalResBlock',
            random_seed: int = DEFAULT_RANDOM_SEED
    ):
        """Residual Block for TemporalConvNet

        :param dilation:
        :param n_filters:
        :param kernel_size:
        :param padding_type:
        :param dropout_rate:
        :param random_seed:
        """
        super(TemporalResidualBlock, self).__init__(name=name)

        # Set the random seed
        self.random_seed = random_seed

        # Validate parameters
        assert dilation in np.logspace(0, 20, num=21, base=2), "Value of 'dilation' parameter must be a power of 2"
        assert padding_type in PERMITTED_PADDING_TYPES, f"Value of 'padding_type' parameter must be one of {PERMITTED_PADDING_TYPES}"
        assert 0. < dropout_rate < 1., r"Value of 'dropout_rate' parameter must lie inside the range (0., 1.)"
        assert final_activation in ['relu', 'linear', 'sigmoid']

        # Gaussian parameter initialisations
        initialiser = tf.keras.initializers.RandomNormal(mean=0.0, stddev=0.01, seed=self.random_seed)

        # First block
        self.conv_1 = layers.Conv1D(
            filters=n_filters,
            kernel_size=kernel_size,
            strides=1,
            padding=padding_type,
            dilation_rate=dilation,
            activation='relu',
            kernel_initializer=initialiser
        )
        self.norm_conv_1 = tfa.layers.WeightNormalization(self.conv_1)
        self.activation_1 = layers.Activation('relu')
        self.dropout_1 = layers.Dropout(rate=dropout_rate, seed=self.random_seed)

        # Second block
        self.conv_2 = layers.Conv1D(
            filters=n_filters,
            kernel_size=kernel_size,
            strides=1,
            padding=padding_type,
            dilation_rate=dilation,
            activation='relu',
            kernel_initializer=initialiser
        )
        self.norm_conv_2 = tfa.layers.WeightNormalization(self.conv_2)
        self.activation_2 = layers.Activation('relu')
        self.dropout_2 = layers.Dropout(rate=dropout_rate, seed=self.random_seed)

        # Correct sizing
        self.shape_regulator = layers.Conv1D(
            filters=n_filters,
            kernel_size=1,
            strides=1,
            padding='same',
            dilation_rate=1,
            activation='relu',
            kernel_initializer=initialiser
        )

        # End-of-block activation (incorporates the skip connection)
        self.final_activation = layers.Activation(final_activation)

    def call(self, inputs, training=None, mask=None):
        initial_inputs = inputs
        x = self.norm_conv_1(inputs)
        x = self.activation_1(x)
        x = self.dropout_1(x) if training else x

        x = self.norm_conv_2(x)
        x = self.activation_2(x)
        x = self.dropout_2(x) if training else x

        if initial_inputs.shape[-1] != x.shape[-1]:
            initial_inputs = self.shape_regulator(initial_inputs)
        assert initial_inputs.shape[-1] == x.shape[-1], f"initial shape {initial_inputs.shape} does not match x shape {x.shape}"

        return self.final_activation(initial_inputs + x)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ptcn']

package_data = \
{'': ['*']}

install_requires = \
['tensorflow-addons>=0.10.0,<0.11.0', 'tensorflow>=2.2.0,<3.0.0']

setup_kwargs = {
    'name': 'ptcn',
    'version': '0.1.2',
    'description': 'TF2 implementation of a Temporal Convolutional Network with a probabilistic twist',
    'long_description': '# ptcn\n\nTensorflow (2.x) implementation of a Temporal Convolutional Network architecture, with a probabilistic twist.\n\nThis project indulges a couple of curiosities:\n\n1. Working with convolutional sequence-to-sequence models a la [An Empirical Evaluation of Generic Convolutional and Recurrent Networks for Sequence Modeling](https://arxiv.org/abs/1803.01271)\n2. Adding a bayesian twist to the network a la [Bayesian Segnet: Model Uncertainty in Deep Convolutional Encoder-Decoder Architectures for Scene Understanding](https://arxiv.org/abs/1511.02680)\n\nThis implementation has been inspired by other projects, including:\n- https://github.com/locuslab/TCN\n- https://github.com/Baichenjia/Tensorflow-TCN\n- https://github.com/philipperemy/keras-tcn\n',
    'author': 'UpstatePedro',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UpstatePedro/ptcn',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
